# coding: utf-8
# This import has to be first for using it on the cluster
import matplotlib as mpl

mpl.use('Agg')

#
# import pip
# def install(package):
#     pip.main(['install', package])
#
# for package in ['numpy','matplotlib','seaborn','ipywidgets','joblib','ipython', 'tqdm', 'futures', 'scipy', 'pandas']:
#     install(package)

import sys
import numpy as np                  # arrays and stuff
import pylab as pl                  # for plotting
import seaborn as sns               # for pretty statistics plotting
from scipy.misc import logsumexp
from scipy.special import erfinv
from scipy.stats import kstest
from scipy.stats import beta
import pandas as pd                 # for reading csv files in convenient format
from parallel import mapp
from shmarray import zeros
import bisect
from collections import Counter
from mpmath import mp
mp.dps = 50


# # Stochastic bouncing ball module

# Next, we set up our model. This is a simple ballistic curve model following the equation:
# $$x(t) = x(0) + v t \cos(\theta)\qquad y(t) = y(0) + v t \sin(\theta) - \frac{1}{2}g t^2$$
#
# where $\theta$ and $v$ are given by the initial velocity vector $\vec{v}_0$as follows:
#
# $$ \theta = \arctan (\vec{v}_0) \qquad v = ||\vec{v}_0||_2 $$
#
# Given the initial velocity vector and position, we can therefore calculate, when the ball hits the ground ($y(t)=0$) the next time. When this happens, we reverse velocity vector's y-coordinate, damp the speed with a factor $\rho$ and add a random perturbance to the resulting angle to model an unregular surface. More precisely, the velocity vector at any point in time is given by:
# $$ \vec{v}(t) = \frac{\partial (x(t),y(t))}{\partial t} = \left(v \cos(\theta), v\sin(\theta) - g t\right)$$
# in particular, we are interested in the next time $t_{n+1}$ of hitting the surface. This time is given by setting the y-coordinate to zero:
# $$ t_{n+1} = \sqrt{\left(\frac{\sin(\theta)v}{g}\right)^2 + \frac{2 y(t_n)}{g}} + \frac{v\sin{\theta}}{g}$$
#
# To bounce, damp and pertube the angle, we simply set the speed and angle at the next time $t_{n+1}$ as follows:
# $$ ||v({t_{n+1}})||_2 = ||\vec{v}(t_n)||_2 \rho \\ \theta(t_{n+1}) = \eta $$
#
# Here, $\eta $ is a random perturbance. For simplicity, we choose a random (uniform) pertubation from a pre-specified list:
# $$\eta \sim \mathcal{U}\{\eta_1,\dots,\eta_m\}$$

# Given this, model, we do some initializations...

# ## Parameters in the model

# In[4]:

# perturbances: sample angles from 10 to 80 degrees relative to the ground (90 degrees would be bouncing on the spot)
eta = np.array([70.,65., 60., 55., 50.]) * (np.pi / 180.)
#eta = np.array([70.,65., 60.]) * (np.pi / 180.)

# gravity
g = 9.81

# initial velocity vector
vecv0 = np.array([10, 10]) / np.sqrt(2)

# initial positional vector (x,y)
pos0 = np.array([0, 0])

# damping factor rho
rho = 0.98

# time begins at zero
t = 0


# Now define the transition for a given time, i.e. update the states (position and velocity). If no time is given, we calculate the time for hitting the surface and update the states accordingly

# ## Implementation of the ball dynamics

# The next two function define the update on the states of the ball between two consequtive hits on the ground and then the actual bouncing mechanis, i.e. perturbing the angle of the velocity vector when hitting the ground.

# In[5]:

def update_state(pos, vel, dt=None):
    bounce = False
    # arctan2 takes first y then x coordinate, hence the -1
    theta = np.arctan2(*(vel[::-1]))
    v = np.linalg.norm(vel)
    if dt is None:
        if np.abs(v) < 1e-09:
            dt = 0.
        else:
            dt = np.sqrt((v * np.sin(theta) / g) ** 2 + pos[1] * 2 / g) + v * np.sin(theta) / g
    veln = np.array([v * np.cos(theta), v * np.sin(theta) - g * dt])
    posn = np.array([pos[0] + v * dt * np.cos(theta), pos[1] + v * dt * np.sin(theta) - 0.5 * g * dt ** 2])
    if posn[1] <= 1e-9:
        # we have hit the surface, therefore reverse y-coordinte
        bounce = True
        veln[1] = np.abs(veln[1])
        # and damp
        veln = veln * rho
    return posn, veln, dt, bounce


def pertube_velocity(vel, states=None, dim=0, choice=None):
    theta = np.arctan2(*(vel[::-1]))
    v = np.linalg.norm(vel)
    # sample new angle for the bounce: uniformly across all possible angles
    if choice is None:
        choice = np.random.randint(len(eta))
    theta = eta[choice]
    newvel = np.array([np.cos(theta), np.sin(theta)]) * v
    prop = 1. / len(eta)
    if states is None:
        states = {'choices': []}
    states['choices'].append(choice)
    return newvel, prop, states


# Sanity check: plot an evolution for a certain discretization

# # Illustration of the bouncing ball (single simulation run)

# In[6]:

dt = 0.005
t = 0.
poss = [pos0]
vels = [vecv0]
ts = [t]
bouncex = [pos0[0]]
it = 0
maxiter = 10000
while (t < 10) and (it < maxiter):
    posn, veln, dt, bounce = update_state(poss[-1], vels[-1], dt=dt)
    if bounce:
        veln, _, _ = pertube_velocity(veln)
        bouncex.append(posn[0])
    poss.append(posn)
    vels.append(veln)
    t += dt
    ts.append(t)
    it += 1

print("Illustration of the bouncing ball")
print("bouncing xses: ", bouncex)
print("diff: ", np.diff(np.array(bouncex)))

sns.set_context('talk')
sns.set_style('ticks')
fig, ax = pl.subplots()
ax.scatter([p[0] for p in poss], [p[1] for p in poss])
ax.set_xlabel('X [m]')
ax.set_ylabel('Height [m]')
pl.xlim(0, 40)
pl.ylim(0, 40)
pl.grid()
sns.despine(offset=10)

# # Setup of the rare event simulation

# Now define the rare event which we would like to analyse its probability of by defining an interval to reach. More precisely, we are interested in the probability that the ball will hit a small range on the surface, defined by the following box, containing an interval for the x-coordinate. When the interval is sufficiently small, the probability of reaching this exact interval is equal to the probability of drawing the exact sequence of angles for each bounce on the surface (given the damping factor is larger than zero).

# In[7]:

rare_event_accuracy = 0.0000001
poss = [pos0]
choice = 0
vels = [vecv0]
ts = [t]
bouncex = [pos0[0]]
nbounces = 7
#nbounces = 4
dones = False
nbounce = 0
while not dones:
    posn, veln, dt, bounce = update_state(poss[-1], vels[-1])
    if bounce:
        veln, _, _ = pertube_velocity(veln, choice=choice)
        bouncex.append(posn[0])
        nbounce +=1
    poss.append(posn)
    vels.append(veln)
    t += dt
    ts.append(t)
    it += 1
    if nbounce == nbounces:
        dones = True

rare_event_pos = poss[-1][0]
rare_event_log_prob = -(float(nbounces) - 1.) * np.log(len(eta))
hitting_box = np.array([rare_event_pos - rare_event_accuracy,
                        rare_event_pos + rare_event_accuracy])
print("--> Determine the rare event")
print("bounces: ", bouncex)
print("rare event position sampled: ", rare_event_pos)
print("Sampled hitbox: ", hitting_box)
print("Log probability of hitting: ", rare_event_log_prob)
print("prob of hitting: ", np.exp(rare_event_log_prob))


# For the simulation setup, we define a functions which simulates a run, with a possible callback of guiding the simulation run. This callback gets as additional argument not only the current velocity-vector to pertube for bouncing, but also an internal state object, which can be updated after the simulation run and is contained within the states object given as input to the simulation routine.

# In[8]:

def simulate_rare_event(states, hitting_box, callback=pertube_velocity, max_bounces=100):
    """Simulate the stochastic bouncing ball using a callback strategy for updating to incorporate a potentially
    different proposal distribution. Returns the sampled points, where the ball hits the surface, the probability of
    the simulated trace (splitted into individual log-factors), the minimal distance to the critical situation over the
    overall trace, and the current state (including a potential internal state of the proposal distribution)"""
    done = False
    pos = states['position']
    vel = states['velocity']
    guiding_state = states['internal sampling states']
    surface_points = []
    # logarithmus of probability
    trace_logp = []
    min_distance = np.inf
    nbounce = 0
    while not done:
        # update the position of the ball according to its physics
        posn, veln, dt, bounce = update_state(pos, vel)
        surface_points.append(posn[0])

        # calculate actual distance to hitting box
        dist = min([np.abs(posn[0] - np.min(hitting_box)), np.abs(posn[0] - np.max(hitting_box))])
        if (posn[0] >= np.min(hitting_box)) and (posn[0] <= np.max(hitting_box)):
            #we hit the box
            min_distance = 0.
        elif dist <= min_distance:
            min_distance = dist
        else:
            pass
        # if we hit the box or jump over it we are ready
        if (posn[0] > np.max(hitting_box)) or (np.abs(dt) < 1e-09) or (nbounce > max_bounces) or (min_distance == 0.):
            done = True
            #announce 'bingo' to z-sampling
            if ('hashmap' in guiding_state):
                # current node will normally be set in guidez but as we break here we must append
                # the node to the list
                guiding_state['hashmap'][guiding_state['current node']]['nvisited'] += 1
                guiding_state['uid trace'].append(guiding_state['current node'])

                if min_distance ==0.:
                    guiding_state['hashmap'][guiding_state['current node']]['bingo'] += 1
        # we the ball will bounce on the floor
        if bounce and (not done):
            # calculate new velocity (pertube_velocity)
            veln, prop, guiding_state = callback(veln, guiding_state, nbounce)
            nbounce += 1
            # prop is the probability of the sample according to the proposal used for sampling
            trace_logp.append(np.log(prop))

        pos = posn
        vel = veln
    states['position'] = pos
    states['velocity'] = vel
    states['internal sampling states'] = guiding_state
    return surface_points, trace_logp, float(min_distance), states


# Now, just a sanity check for testing the simulation method

# In[9]:

sp, tlp, h, states = simulate_rare_event({'position': pos0,
                                          'velocity': vecv0,
                                          'internal sampling states': None},
                                         hitting_box)
print("--> sanity check: simulate rare event")
print("surface points: ", sp)
print("log prop: ", tlp)
print("dist to hitbox: ", h)
print("states: ", states)
print("sum of choices: ", sum(states['internal sampling states']['choices']))
print("----------------------------------------------------------------------------------------------------")
sp, tlp, h, states = simulate_rare_event({'position': pos0,
                                          'velocity': vecv0,
                                          'internal sampling states': None},
                                         hitting_box)
print("--> sanity check: simulate rare event")
print("surface points: ", sp)
print("log prop: ", tlp)
print("dist to hitbox: ", h)
print("states: ", states)
print("sum of choices: ", sum(states['internal sampling states']['choices']))
print("----------------------------------------------------------------------------------------------------")

# # Setup parallel execution and progress bars

# Set up the parallel computation with progress bar etc. For this we make some definitions (mainly copied from the internets)

# In[10]:

from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed
from numpy.random import seed as np_seed
from random import seed as py_seed
#import copy_reg
#import types
import time
#remove next line when ready
#tqdm=lambda x:x

# def _pickle_method(method):
#    attached_object = method.im_self or method.im_class
#    func_name = method.im_func.func_name##

#    if func_name.startswith('__'):
#        func_name = filter(lambda method_name: method_name.startswith('_') and method_name.endswith(func_name), dir(attached_object))[0]

#   return (getattr, (attached_object, func_name))

# copy_reg.pickle(types.MethodType, _pickle_method)


def parallel_process(array, function, n_jobs=16, use_kwargs=False, front_num=3):
    """
        A parallel version of the map function with a progress bar.

        Args:
            array (array-like): An array to iterate over.
            function (function): A python function to apply to the elements of array
            n_jobs (int, default=16): The number of cores to use
            use_kwargs (boolean, default=False): Whether to consider the elements of array as dictionaries of
                keyword arguments to function
            front_num (int, default=3): The number of iterations to run serially before kicking off the parallel job.
                Useful for catching bugs
        Returns:
            [function(array[0]), function(array[1]), ...]
    """
    # We run the first few iterations serially to catch bugs
    if front_num > 0:
        front = [function(**a) if use_kwargs else function(a) for a in array[:front_num]]
    # If we set n_jobs to 1, just run a list comprehension. This is useful for benchmarking and debugging.
    if n_jobs == 1:
        return front + [function(**a) if use_kwargs else function(a) for a in tqdm(array[front_num:])]
    # Assemble the workers
    with ProcessPoolExecutor(max_workers=n_jobs) as pool:
        # Pass the elements of array into function
        if use_kwargs:
            futures = []
            for a in array[front_num:]:
                max_rand = 4294967295
                rseed = np.random.randint(0, max_rand - 1)
                rseed = rseed % 4294967295

                def f():
                    np_seed(rseed)
                    py_seed(rseed)
                    return function(**a)

                futures.append(pool.submit(f))
                time.sleep(0.01)
                # futures = [pool.submit(function, **a) for a in array[front_num:]]
        else:
            futures = []
            for a in array[front_num:]:
                max_rand = 4294967295
                rseed = np.random.randint(0, max_rand - 1)
                rseed = rseed % 4294967295

                def f():
                    np_seed(rseed)
                    py_seed(rseed)
                    return function(a)

                futures.append(pool.submit(f))
                # futures.append(pool.submit(function, a))
                time.sleep(0.01)
                #            futures = [pool.submit(function, a) for a in array[front_num:]]
        kwargs = {
            'total': len(futures),
            'unit': 'it',
            'unit_scale': True,
            'leave': True
        }
        # Print out the progress as tasks complete
        for f in tqdm(as_completed(futures), **kwargs):
            pass
    out = []
    # Get the results from the futures.
    for i, future in tqdm(enumerate(futures)):
        try:
            out.append(future.result())
        except Exception as e:
            out.append(e)
    return front + out


# ## Testing the parallel execution with the naive estimation of the rare event probability

# In[11]:

## Estimating the probability of hitting the small box using naive sampling, i.e. without any guiding
nsamples = 20000
initstate = {'position': pos0, 'velocity': vecv0, 'internal sampling states': None}
min_distances = zeros(nsamples)
from copy import deepcopy as cp


def f(n):
    sp, tlp, d, h = simulate_rare_event(cp(initstate), hitting_box)
    min_distances[n] = d
    # return d


# min_distances = parallel_process(range(nsamples), f, n_jobs=20)
# min_distances = map(f, range(nsamples))

#map(f, range(nsamples))
#min_distances = np.array(min_distances)
#print("Estimated probability of hitting: ", np.sum(1 * min_distances < 1e-09) / float(nsamples))


# # Guiding strategies

# ## Importance sampling strategy with product approximation

# Now, we implement a first guiding strategy, based on adaptive importance sampling using a product approximation as proposal distribution, which is updated using the cross-entropy method (see https://scholar.google.com/citations?view_op=view_citation&hl=en&user=g00R4hoAAAAJ&citation_for_view=g00R4hoAAAAJ:tSVUDDkujAIC).
#
# Each bouncing of the ball defines a new dimension for which the las vegas strategy stores a probability vector.
# After a simulation run, this probability vector is updated by increasing or decreasing the probability of the simulation run, depending on, if it is considered to be a desireable simulation run, or not.

# In[12]:
    
def product_approximation_proposal_guide(vel, state, dim):
    init_factor = 1.
    if state is None:
        # if we have not sampled yet, we have to set up the proposal as
        # independent (between time-points) distribution
        ntab = len(eta)  # number of possibilities to sample from = dimensions of a multinomial distribution
        ndim = 1  # number of dimensions = number of time-points already seen, i.e. initially 1
        state = {'proposal': np.ones((ntab, ndim)) * init_factor / float(ntab),
                 'choices': []}
    qprop = state['proposal']
    ntab, ndim = qprop.shape
    choices = state['choices']

    if dim > (ndim - 1):
        # if we need to sample for times larger than the ones already sampled, we add another factor
        qprop = np.hstack((qprop, np.ones((ntab, 1)) * init_factor))
    # draw a sample from the current dimension using multinomial distribution
    # here, we also take care of the normalization, although it should sum up to one
    # in that case, normalizing doesn't change anything
    sample = np.random.multinomial(1, qprop[:, dim] / np.sum(qprop[:, dim]))
    rand_ind = np.array(range(len(eta)))[sample > 0]
    # select the corresponding perturbation
    angle_perturb = eta[rand_ind]

    # probability of this choice
    prop = (qprop[:, dim] / np.sum(qprop[:, dim]))[rand_ind]
    # remember choice for next time in the format (choice, dimension)
    choices.append((rand_ind[0], dim))
    # arctan 2 reverses the roles of y and x coordinates
    theta = np.arctan2(*(vel[::-1]))
    # remember speed of the ball, as we just want to sample new angle
    v = np.linalg.norm(vel)
    # set new anlge
    theta = angle_perturb
    # construct the new velocity vector from angle and speed
    newvel = np.array([np.cos(theta), np.sin(theta)]) * v
    # remember new (added dimensions) proposal distribution
    state['proposal'] = qprop
    state['choices'] = choices
    return newvel, prop, state

def independent_exp_proposal_guide(vel, state, dim):
    """Sample independent from the dimension(time) from the same distribution"""
    if state is None:
        state = {'proposal': np.ones(len(eta))/float(len(eta)),
                 'choices': [],
                 'one trace': []}
        # we have not used the guiding yet, therefore, we have to set it up
    elif (not 'one trace' in state.keys()):
        state['one trace'] = []
    elif (not 'choices' in state.keys()):
        state['choices'] = []
    else:
        pass
    prop = state['proposal']
    sample = np.random.multinomial(1, prop)
    rand_ind = np.array(range(len(eta)))[sample > 0]
    angle_perturb = eta[rand_ind]
    pt = prop[rand_ind]
    #state['choices'].append(rand_ind[0])
    rand_choice = rand_ind[0]
    state['one trace'].append(rand_choice)
    state['choices'].append(rand_choice)
    theta = np.arctan2(*(vel[::-1]))
    # remember speed of the ball, as we just want to sample new angle
    v = np.linalg.norm(vel)
    # set new anlge
    theta = angle_perturb
    # construct the new velocity vector from angle and speed
    newvel = np.array([np.cos(theta), np.sin(theta)]) * v
    # remember new (added dimensions) proposal distribution
    return newvel, pt, state


def cross_entropy_update_independent(state, feedback, quantile, batchsize):
    # no feedback seen so far, so remember them in criticalities and also remember traces
    if not ('criticalities' in state.keys()):
        state['criticalities'] = [[feedback,0]]
        state['choicesfortrace'] = [state['one trace']]
        state['ntraces'] = 0
        state['minlen'] = len(state['one trace'])
        state['one trace'] = []
        state['startp'] = state['proposal'][choice]
        return state
    # we have seen criticalities but not enough, remember the same
    if len(state['criticalities']) < batchsize:
        state['ntraces'] +=1
        # criticalities are sorted for each feedback, we know the corresponding id of the trace
        bisect.insort(state['criticalities'], [feedback, state['ntraces']])
        # here the trace-choices are remembered
        state['choicesfortrace'].append(cp(state['one trace']))
        # store the minimal length of a trace
        state['minlen'] = min([len(state['one trace']),state['minlen']])
        state['one trace'] = []
        return state
    else:
        state['ntraces'] +=1
        bisect.insort(state['criticalities'], [feedback, state['ntraces']])
        state['choicesfortrace'].append(cp(state['one trace']))
        state['minlen'] = min([len(state['one trace']), state['minlen']])
        state['choices'] = []
    if 'gamma' in state.keys():
        gamma = state['gamma']
    else:
        gamma = np.inf
    ind = int(np.floor(batchsize * quantile))
    # the quantile feedback, i.e. the quantile-most critical feedback value
    gs = state['criticalities'][ind][0]
    if gs <= gamma :
        state['gamma'] = gs
        # select all traces, which are at least as critical as gs
        trace_indices = np.array(state['criticalities'][:ind])[:,1]
        L = []
        for ti in trace_indices:
            L = L + list(np.array(state['choicesfortrace'])[int(ti)])
        # how often have we seen which choice
        counter = Counter(L)

        p = np.zeros(state['proposal'].shape)
        for i,v in zip(counter.keys(), counter.values()):
            # calculate mean of p*(I_<gamma) * (1./q)
            p[i] = float(v)/float(len(trace_indices) * len(state['choicesfortrace'][0])) * (1./float(len(eta))) * (1./ (state['proposal'][i]))
        # calcualte exponential family with p as parameter / sufficient statistics
        p = np.exp(p)/np.sum(np.exp(p))
        #print("\rDelta to start choice p (+ is good): "+str(p[choice]-state['startp']),)
        state['proposal'] = p
    state.pop('criticalities',None)
    state.pop('choicesfortrace', None)
    state.pop('ntraces', None)
    state['one trace'] = []
    state.pop('minlen', None)
    return state


def cross_entropy_update(state, feedback, quantile=0.2, batchsize=30, learning_rate=0.1):
    # we haven't see enough samples, so no updates  yet
    # if we haven't seen anything, we add criticalities to keys and add feedback for logging
    if not ('criticalities' in state.keys()):
        state['criticalities'] = [feedback]
        state['Xs'] = [state['choices']]
        return state
    # if it in keys, it could still be empty, so we start filling it
    if len(state['criticalities']) == 0:
        state['criticalities'] = [feedback]
        state['Xs'] = [state['choices']]
        return state
    # we have seen some states, but not enough to make an update
    if len(state['criticalities']) < batchsize:
        state['Xs'].append(state['choices'])
        state['criticalities'].append(feedback)
        return state
    # now we have at least batchsize number of samples, so we do an update
    lporig = lambda x: -np.log(len(eta)) * np.prod(x.shape)
    if 'gamma' in state.keys():
        gamma = state['gamma']
    else:
        gamma = np.inf
    gs = np.array(state['criticalities'])
    gs.sort()
    # gamma marks the point below which we collect all traces as being used for updating
    gam = gs[int(np.floor((len(gs) - 1) * (quantile)))]
    if gam < gamma:
        gamma = gam
    else:
        #if the new quantile isn't more critical than the last one, no update (and also no batch-cleaning)
        return state
    # select the samples with being at least as close as the quantile
    ind = np.array(state['criticalities']).flatten() < gamma
    # 'Xs' contains the choices, as tuple (index, dimension(time))
    cs = state['Xs']

    # this will contain the new optimized proposal
    op_state = np.zeros(state['proposal'].shape)

    # go through all selected (sufficiently critical) traces
    for k in np.array(range(len(cs)))[ind]:
        for a in np.array(cs[k]):
            op_state[a[0], a[1]] += (1./float(len(ind))) * min([(1. / state['proposal'][a[0], a[1]]), 3]) * (1./ float(len(eta)))

   # op_state[np.array(cs)[ind][:,:,0], np.array(cs)[ind][:,:,1]] += (1./float(len(ind))) * state['proposal']
    op_state = np.exp(op_state) / np.sum(np.exp(op_state),axis=0)
    #for i in range(op_state.shape[1]):
    #    op_state[:, i] = op_state[:,i] / np.sum(op)
            #np.sqrt(op_state[:, i]) / np.sum(np.sqrt(op_state[:, 1]))
    prop = op_state  # state['proposal'] + learning_rate*(op_state - state['proposal'])
    #for i in range(prop.shape[1]):
    #    prop[:, i] = prop[:, i] / np.sum(prop[:, i])
#        print("Proposal for choice of choice: " + str(prop[0,0]))
    state['proposal'] = prop
    state['criticalities'] = []
    state['gamma'] = gamma
    return state


def update_las_vegas_state(state, feedback):
    """
    updates the proposal distribution within the state dictionary based on the given feedback.
    for the update, we use the cross-entropy method
    here: it is assumed that the feedback is a float indicating:
    feedback ~ 0 : last simulation run had the desired property
    feedback ~ 1 : lsat simulation run was quite far away from the desired prop
    """
    choices = state['choices']
    prop = state['proposal']
    for i, dim in choices:
        if feedback > 0.005:
            prop[i, dim] = prop[i, dim] * 0.99
        else:
            prop[i, dim] = prop[i, dim] * 1.01
    state['proposal'] = prop
    state['choices'] = []
    return state


## defining sampling strategy by defining a callback and an update function. for the naive
## strategy, the update does nothing but return the state of the guiding function, which
## in the naive strategy-case is always none

def init_crossentropy_strategy(batchsize=30, quantile=0.2):
    crossentropy_strategy = {'callback': product_approximation_proposal_guide,
                             'update': lambda state, feed: cross_entropy_update(state, feed, quantile=quantile, batchsize=batchsize),
                             'name': 'AIS w CE (batch=%d,q=%f)'%(batchsize,quantile)}
    return crossentropy_strategy

def init_cross_entropy_independent_strategy(batchsize=30, quantile=0.2):
    strategy = {'callback': independent_exp_proposal_guide,
                'update': lambda state, feed: cross_entropy_update_independent(state, feed, quantile=quantile, batchsize=batchsize),
                'name': 'AIS w CE ind. (batch=%d,q=%f)'%(batchsize,quantile)}
    return strategy

def naive_update(state, feedback):
    state['choices'] = []
    return state


naive_strategy = {'callback': pertube_velocity,
                  'update': naive_update,
                  'name': 'Direct sampling'}


# ## Guided simulation a la z-score

# In[13]:

def guide_samplez(vel, state, dim):
    """
    @param vel: velocity vector (2-d numpy array of x and y coordintes)
    @param state: internal state of the guiding strategy
    @param dim: it's the dim-time to hit the floor
    """

    # probability distribution according to the original model
    porig = np.ones(len(eta)) / float(len(eta))

    if state is None:
        # create root node
        state = {'choices': [],
                 'current node': 0,
                 'uids': [0],
                 'hashmap': {0: {'nvisited': 0,
                                 'porig': 1. / float(len(eta)),
                                 'qprop': 1. / float(len(eta)),
                                 'zvalues': [],
                                 'criticalities': [],
                                 'first prop': 0,
                                 'bingo': 0}},
                 'uid trace': []}
    if dim == 0:
        state['current node'] = 0
        state['choices'] = []

    current_node = state['current node']
    hashmap = state['hashmap']
    nvisited = hashmap[current_node]['nvisited']
    #print("Curr: " , state['current node'], "besucht: ",nvisited, end=" ")
    if nvisited == 0:
        children = []
        for i in range(len(eta)):
            node_id = state['uids'][-1] + 1
            state['uids'].append(node_id)
            hashmap[node_id] = {'nvisited': 0,
                                'porig': 1. / float(len(eta)),
                                'qprop': 1. / float(len(eta)),
                                'criticalities': [],
                                'zvalues': [],
                                'angle-id': i,
                                'first prop': 0,
                                'parent node': current_node,
                                'bingo': 0}
            children.append(node_id)
        hashmap[current_node]['children'] = children
        theta = np.arctan2(*(vel[::-1]))
        v = np.linalg.norm(vel)
        # sample new angle for the bounce: uniformly across all possible angles
        random_child = np.random.randint(len(eta))
        theta = eta[random_child]
        newvel = np.array([np.cos(theta), np.sin(theta)]) * v
        hashmap[current_node]['nvisited'] = hashmap[current_node]['nvisited'] + 1
        prop = hashmap[hashmap[current_node]['children'][random_child]]['porig']
        hashmap[current_node]['first prop'] = prop
        state['current node'] = hashmap[current_node]['children'][random_child]

        # save the angle-id in list of choices and uid
        state['choices'].append(state['hashmap'][state['current node']]['angle-id'])
        state['uid trace'].append(current_node)
        state['hashmap'] = hashmap
        # print("Chld: " ,hashmap[current_node]['children'], end=" ")
        # print("NewI " , state['current node'], "besucht: ", hashmap[state['current node']]['nvisited'])

        # as we havent visited the node, we haven't ignored anything
        pignored = 0.

    else:
        # print("Chld: " ,hashmap[current_node]['children'], end=" ")


        hashmap[current_node]['nvisited'] = hashmap[current_node]['nvisited'] + 1

        #iterate over childs; if one child has never visited chose it, else draw using weights
        nextNode = -1
        for e in hashmap[current_node]['children']:
            if hashmap[e]['nvisited'] == 0:
                nextNode = e
                break
        if nextNode != -1:
            prop = hashmap[state['current node']]['first prop']
            state['current node'] = nextNode
            theta = eta[hashmap[nextNode]['angle-id']]
            v = np.linalg.norm(vel)
            newvel = np.array([np.cos(theta), np.sin(theta)]) * v

        else:
            # collect all proposal probabilities for all children of the current node
            qprop = np.array([hashmap[e]['qprop'] for e in hashmap[current_node]['children']])

            # ensure that the probabilities sum to one
            qprop = qprop / np.sum(qprop)

            # draw a single sample from the proposal distribution
            sample = np.random.multinomial(1, qprop)

            # extract the index of the sample (which has length of the qprop vector
            random_child = np.array(range(len(eta)))[sample > 0][0]

            # proposal probability of the sampled child
            prop = qprop[random_child]

            # print("random child: " , random_child)
            theta = eta[random_child]
            v = np.linalg.norm(vel)
            newvel = np.array([np.cos(theta), np.sin(theta)]) * v
            # set sampled child as new current node
            state['current node'] = hashmap[current_node]['children'][int(random_child)]
        state['uid trace'].append(current_node)
        state['choices'].append(state['hashmap'][state['current node']]['angle-id'])
        state['hashmap'] = hashmap
        #print("NewE: " , state['current node'], "besucht: ", hashmap[state['current node']]['nvisited'],
        # "Z: ",hashmap[state['current node']]['zvalue'])

    #now correct for it so that we can do sampling without replacement using q
    return newvel, prop, state



def update_samplez(state, feedback, tau=0.01):
    uid_trace = state['uid trace']
    hashmap = state['hashmap']
    parent = 0
    for node in uid_trace:
        hashmap[node]['criticalities'].append(feedback)
        if len(hashmap[node]['criticalities']) >= 2:
            # get minimum visits from all children
            nmin = min([hashmap[e]['nvisited'] for e in hashmap[parent]['children']])
            cs = np.array(hashmap[node]['criticalities'])
            z = (tau - cs.mean()) / (cs.std() + 1e-5)
            hashmap[node]['zvalues'].append(z)
            f = 0.5 + 0.5 * nmin
            if np.abs(z) > 1e3:
                z = 1e3
            if f > 50.:
                f = 50.
            w = 1. / (np.abs(z) + 1.) ** (f)
            hashmap[node]['qprop'] = w
        parent = node
    state['hashmap'] = hashmap
    state['uid trace'] = []
    return state


def init_z_strategy():
    z_strategy = {'callback': lambda v,s,d: guide_samplez(vel=v, state=s, dim=d),
                  'update': update_samplez,
                  'name': 'Z-guiding'}
    return z_strategy


# ## Testing guiding simulation

# In the next cell, we compare the las-vegas strategy with the naive strategy

# In[14]:

def guided_simulation(nsamples, strategy, batchsize=1):

    feedback = np.zeros(nsamples)
    logp_orig = np.zeros(nsamples)
    logq_prop = np.zeros(nsamples)
    log_eta = np.zeros(len(eta))
    d_hitbox = np.zeros(nsamples)
    max_distance = 1.  # np.max(hitting_box)/20.
    initstate = {'position': pos0,
                 'velocity': vecv0,
                 'internal sampling states': None,
                 'node ids': None}
    state = cp(initstate)
    for n in range(nsamples):
        state['position'] = pos0
        state['velocity'] = vecv0

        # logqprop is on trace of logarithmus of qprop
        sp, logqprop, h, states = simulate_rare_event(state,
                                                      hitting_box,
                                                      callback=strategy['callback'])
        # save the distance to hole
        d_hitbox[n] = h
        #feed = h / max_distance
        feedback[n] = h / max_distance
        # take the sum of the eta indices because to hit the hole you must select always the first
        feedback[n] = sum(state['internal sampling states']['choices'])
        for choise in state['internal sampling states']['choices']:
            log_eta[choise] += 1

        #if (sum(state['internal sampling states']['choices']) < 1):
        #    print('feedback h: ', h)
        #    print('Strat: ', strategy['name'])
        #    print('State: ', state['internal sampling states']['hashmap'][state['internal sampling states']['current node']])

        var_sum = 0
        if 'AIS' in strategy['name']:
            var_sum = np.sum(np.array(states['internal sampling states']['one trace']))
            counter = Counter(states['internal sampling states']['one trace'])
            for i,v in zip(counter.keys(), counter.values()):
                log_eta[i] +=v
            feedback[n] = var_sum
            #feedback[n] = h / max_distance

        if ((h < 1e-08) and (var_sum>0)) and strategy['name']=='AIS w CE ind. (batch=100,q=0.200000)':
            print('feedback h: ',h)
            # print('sum :', sum)
        # if (h < 1e-08) and (feedback[n] > 0):
        #      print('feedback h: ', h)
        #      print('sum: ', feedback[n], ' etas: ', state['internal sampling states']['choices'])
        #      print('Strat: ', strategy['name'])
        logq_prop[n] = np.sum(np.array(logqprop))
        logp_orig[n] = -np.log(len(eta)) * len(logqprop)
        gstate = state['internal sampling states']

        # update the proposal (tree with the feedback, etc. )
        gstate = strategy['update'](gstate, feedback[n])
        state['internal sampling states'] = gstate
    return feedback, logp_orig, logq_prop, state, log_eta, d_hitbox


def estimate_rare_event_prop(feedback, logp, logq, threshold=1e-09):
    indices = np.array(range(len(feedback)))[feedback < threshold]
    if len(indices) == 0:
        return 0.
    lp = logp[indices]
    lq = logq[indices]
    # if proposal and actual are close, we assume naive sampling
    if np.abs(np.max(lp - lq)) < 1e-09:
        phat = float(len(indices)) / float(len(feedback))
    else:
        phat = np.exp(logsumexp(lp - lq) - np.log(float(len(logp))))
    return phat


def test_strategies(strategies, nsamples=500):
    # to prevent, that numerical warnings have no effect
    np.seterr(all='raise')


    for strat in strategies:
        name = strat['name']
        start =  time.time()
        f, lp, lq, state, leta, dhbox = guided_simulation(500, strat)
        end = time.time()
        phat = estimate_rare_event_prop(f, lp, lq, threshold=0.000001)
        print("=" * 80)
        print("Strategy: ", strat['name'])
        print("Estimated prob: ", phat)
        print("True rare event prop: ", np.exp(rare_event_log_prob))
        print("Computation took %0.3f s" % ((end-start)))
        if name == "AIS w cross-entropy":
            print("states of proposal: ", state['internal sampling states']['proposal'])
            # if name == "Z-guiding":
            #    print(" states of proposal ", state)


# # Evaluation

# ## Confidence interval calcualtions

# ### Binomial confidence interval

# For the evaluation, we need to calculate confidence intervals for the different estimators. We start with the simple binomial confidence interval for the naive estimator. This is also known as the Clopper-Pearson confidence interval, or exact confidence interval. This depend on the total amount of samples (nall), the number of successes (nsucc) and the desired confidence level (specified as the error-probability $\alpha$, i.e. a confidence of $1-\alpha$).

# In[15]:

def clopper_pearson_conf(nsucc, nall, alpha=0.05):
    """calculates the confidence interval based on the number of successes
    and overall trials using the clopper pearson exact confidence interval"""
    lo = beta.ppf(alpha / 2., nsucc, nall - nsucc + 1)
    hi = beta.ppf(1. - alpha / 2., nsucc + 1, nall - nsucc)
    if nsucc == 0:
        lo = 0.
        hi = 1. - (alpha / 2.) ** (1. / nall)
    return np.array([lo, hi])


# ### Asymptotic Gaussian confidence intervals

# For a self-normalizing importance sampling estimate, which is assumed to be given by the following formula:
#
# $$ \hat{p} = \frac{ \sum_{i=1}^N f(x_i) \frac{p(x_i)}{q(x_i)} } {\sum_{i=1}^N \frac{p(x_i)}{q(x_i)}} $$
#
# For this estimator, we have the following expression for the variance:
#
# $$ \sigma(\hat{p})^2 = \sum_{i=1}^N w_i^2 (f- \hat{p})^2 $$
#
# where
# $$ w_i = \frac{p(x_i)/q(x_i)}{\sum_i p(x_i)/q(x_i)} $$
# are the normalized importance weights. Assuming that the number of samples is sufficiently large and the weights are hence approximately Gausian distributed, we have the following Gaussian confidence interval:
#
# $$ C(\alpha) = \hat{p} \pm \sqrt{\frac{2\sigma(\hat{p})}{N}} \Phi^{-1}(\alpha)$$
#
# where $\Phi^{-1}$ is the inverse of the error-fnuction.
#

# In[16]:

def asymptotic_gaussian_confidence_interval(logweights, funvalues, alpha):
    pindex = (funvalues > 0.01)
    if sum(1 * (pindex)) == 0:
        return np.array([0., 0.])
    logw = logweights - logsumexp(logweights)
    phat = np.exp(logsumexp(logw[pindex]))
    sigma = np.exp(logsumexp(2. * logw + 2. * np.log(np.abs(1. * (funvalues > 0.01) - phat))))
    err = np.sqrt(sigma / len(logweights)) * (np.sqrt(2.) * erfinv(alpha))
    lo = phat - err
    hi = phat + err
    return np.array([lo, hi])


# ### Bootstrap confidence intervals

# The above confidence interval calculation assumes that the number of samples has reached the asymptotic regime, where they are approximately Gaussian. Alternatively, we could calculate a bootstrap confidence interval.

# In[17]:

def bootstrap_confidence_interval(logweights,
                                  fvalues,
                                  alpha=0.05, nbootstrapsamples=100):
    N = len(logweights)
    empirical_result_distribution = zeros(nbootstrapsamples)

    def f(n):
        indices = np.random.randint(N, size=N)
        lw = logweights[indices]
        fv = fvalues[indices]
        findices = np.array(range(len(fvalues)))[fvalues > 0.01]
        if len(findices) == 0:
            estimate = 0.
        else:
            lp = lw[findices]
            estimate = np.exp(logsumexp(lp) - logsumexp(lw))
        # estimate = estimate_rare_event_prop(fv, logp=lw, logq=np.zeros(len(lw)), threshold=0.5)
        empirical_result_distribution[n] = estimate
        # return estimate

    # empirical_result_distribution = parallel_process(range(nbootstrapsamples),
    #                                                 f,
    #                                                 n_jobs=20)
    mapp(f, range(nbootstrapsamples))
)
    emp = np.array(empirical_result_distribution)
    emp.sort()
    conf = np.array([emp[int(np.floor((alpha / 2.) * nbootstrapsamples))],
                     emp[min([int(np.ceil((1. - alpha / 2.) * nbootstrapsamples)), nbootstrapsamples-1])]])
    return conf


# # Comparison of guiding strategies

# Here, we compare the guiding strategies we have defined before. For now, these include the direct sampling (monte carlo) and the adaptive importance sampling, called las-vegas, which makes a product approximation to the sampling distribution and updates it iteratively. To compare guiding strategies, we plot a CDF $\pm$ a confidence interval. Here the cdf computes the (estimated) probability of generating samples, which are at least as close to the hitting box as a specified threshold. The estimated probability is then plotted as a function of the given threshold.

# In[18]:

def compare_strategies_single_run(strategies=None, fn_plot=None, nsamples=5000, ax =None):
    if strategies is None:
        strategies = [naive_strategy, crossentropy_strategy, z_strategy]
    if fn_plot is None:
        fn_plot = './cdf_intervals_comparison'
    thresholds = np.linspace(0, 0.1, 10)
    confidence_level = 0.99
    save_plot = True
    if ax is None:
        sns.set_context('talk')
        sns.set_style('ticks')
        fig, ax = pl.subplots()
    else:
        save_plot=False
    ax.set_yscale('log')

    import itertools

    palette = itertools.cycle(sns.color_palette())
    color = next(palette)
    for strat in strategies:
        fvalues, lp, lq, state, leta, dhbox = guided_simulation(nsamples=nsamples, strategy=strat)
        conf_interv_d = []  # direct / clopper pearson
        conf_interv_a = []  # asymptotic gaussian
        conf_interv_b = []  # bootstrap

        for it, threshold in enumerate(thresholds):
            phat = estimate_rare_event_prop(fvalues, lp, lq, threshold=threshold)
            # print("Strategy: ", strat['name'], "  estimate: ", phat)
            if np.max(np.abs(lp - lq)) < 1e-09:
                conf_interval = clopper_pearson_conf(nsucc=np.sum((fvalues < threshold) * 1),
                                                     nall=len(fvalues),
                                                     alpha=1. - confidence_level)
                conf_interv_d.append(conf_interval)
            else:
                conf_interval_gauss = asymptotic_gaussian_confidence_interval(logweights=(lp - lq),
                                                                              funvalues=1 * (fvalues < threshold),
                                                                              alpha=1. - confidence_level)
                conf_interv_a.append(conf_interval_gauss)
                # print("asymptotic gaussian: ", conf_interval_gauss)

                samples = np.exp(lp - lq - logsumexp(lp - lq)) * (fvalues < threshold)
                conf_interval_boot = bootstrap_confidence_interval(logweights=(lp - lq),
                                                                   fvalues=1 * (fvalues < threshold),
                                                                   alpha=1. - confidence_level,
                                                                   nbootstrapsamples=10000)
                conf_interv_b.append(conf_interval_boot)
                # print(" bootstrap: ", conf_interval_boot)
        if len(conf_interv_d) > 0:
            conf_interv_d = np.array(conf_interv_d)
            ax.fill_between(thresholds, conf_interv_d[:, 0], conf_interv_d[:, 1], facecolor=color, alpha=0.5,
                            label=r'Clopper pearson $\alpha$= %g' % (1. - confidence_level))
            ax.plot(thresholds, np.mean(conf_interv_d, axis=1), color=color, label='Direct sampling')
            color = next(palette)
        if (len(conf_interv_a) > 0):
            conf_interv_a = np.array(conf_interv_a)
            ax.fill_between(thresholds, conf_interv_a[:, 0], conf_interv_a[:, 1], facecolor=color, alpha=0.5,
                            label=r'Asympotic Gaussian $\alpha$= %g' % (1. - confidence_level) + ' for '+ strat['name'])
            ax.plot(thresholds, np.mean(conf_interv_a, axis=1), color=color, ls='--',
                    lw=5, label='direct mean for ' + strat['name'])
            color = next(palette)
        if (len(conf_interv_b) > 0):
            conf_interv_b = np.array(conf_interv_b)
            ax.fill_between(thresholds, conf_interv_b[:, 0], conf_interv_a[:, 1], facecolor=color, alpha=0.5,
                            label=r'Bootstrap $\alpha$= %g' % (1. - confidence_level) + ' for '+ strat['name'])
            ax.plot(thresholds, np.mean(conf_interv_b, axis=1), color=color,
                    ls='-.', label='Bootstrap mean for ' + strat['name'], lw=2)
            color = next(palette)

    pl.legend()
    ax.set_xlabel('Minimal distance [m]')
    ax.set_ylabel(r'$P($distance$\leq x )$')
    sns.despine(offset=10, ax=ax)
    if save_plot:
        fig.savefig(fn_plot+ '.pdf', bbox_inches='tight')


def clopper_pearson_conf(nsucc, nall, alpha=0.05):
    """calculates the confidence interval based on the number of successes
    and overall trials using the clopper pearson exact confidence interval"""
    lo = beta.ppf(alpha/2., nsucc, nall-nsucc+1)
    hi = beta.ppf(1. - alpha/2., nsucc+1, nall-nsucc)
    if nsucc ==0:
        lo = 0.
        hi = 1.- (alpha/2.)**(1./nall)
    return np.array([lo,hi])

def logsumexpf(lw):
    la = np.max(lw)
    return np.sum(np.array(list(map(mp.exp,lw-la)))) + la

def compute_bootstrap_confidence_interval(fs, logporig, logqprop, confidence_level=0.9, nbootstrap=1000):
    resampled = np.zeros(nbootstrap)
    for n in tqdm(range(nbootstrap), 'bootstrapping...'):
        if nbootstrap==1:
            ind = range(len(logporig))
        else:
            ind = np.random.randint(len(logporig), size=len(logporig))
        lws = np.array(list(map(lambda x: mp.mpf(x), logporig - logqprop)))
        f = fs[ind]
        lws = lws[ind]
        if len(f[f>0.001])>0:
            resampled[n] = mp.exp(logsumexpf(lws[f>0.01]) - logsumexpf(lws))
        r = resampled[n]

    resampled.sort()
    return np.array([resampled[int(np.floor((1.-confidence_level)*nbootstrap))],
                    resampled[min([int(np.ceil((confidence_level)*nbootstrap)), nbootstrap-1])]])




from graphviz import Digraph

# print the guided tree using graphiz
def plotTree(data, nsamples, nrepetitions):

    tau = 0.01


    for index, item in enumerate(data):
        dot = Digraph(comment='Z-Guided-Tree', \
                      format='PDF', \
                      node_attr={'labelfontsize': '8.0', 'labelfontname': 'Arial'}, \
                      graph_attr={'rankdir': 'LR'})

        #dot.attr(size='8,5')
        #dot.graph_attr['rankdir'] = 'LR'
        #dot.graph_attr['labelfontsize'] = '8.0'

        nodedict = item['internal sampling states']['hashmap']


        for parent, children in nodedict.items():
            z = -1
            min_crit = -1
            if len(children['criticalities']) >= 2:
                cs = np.array(children['criticalities'])
                z = (tau - cs.mean()) / (cs.std() + 1e-5)
            if len(children['criticalities']) > 0:
                # min criticality
                min_crit = min(children['criticalities'])
            if children['nvisited'] > 0:
                if children['bingo'] > 0:
                    dot.node(str(parent), \
                             str(parent) + ' - #' + \
                             str(children['nvisited']) + '\n p:' + \
                             str(round(children['porig'],5)) + '\n q:'+ \
                             str('{:0.3e}'.format(children['qprop'])) + '\n z:'+ \
                             str(round(z,5)) + ' '+ str(round(cs.mean(),5)) + ' '+str(round(cs.std(),5)) + '\n d:' + \
                             str(round(min_crit,5)), \
                             color='green', labelfontname='Courier')
                else:
                    dot.node(str(parent), \
                             str(parent) + ' -  #' + \
                             str(children['nvisited']) + '\n p:' + \
                             str(round(children['porig'],5)) + '\n q:'+ \
                             str('{:0.3e}'.format(children['qprop'])) + '\n z:'+ \
                             str(round(z, 5)) + ' ' + str(round(cs.mean(), 5)) + ' ' + str(round(cs.std(), 5)) + '\n d:' + \
                             str(round(min_crit,5)), \
                             color='red')
            else:
                dot.node(str(parent), str(parent), labelfontname='Arial', labelfontsize='22')
            # draw the edges
            if ('children' in children):
                for child in children['children']:
                    dot.edge(str(parent),  str(child))

        #print(dot.source)
        dot.render('test-output/guided-tree_nsamples_%d_nrepetitions_%d_#%d.gv'%(nsamples,nrepetitions,index+1), view=False)
        #dot.format='SVG'
        #dot.render('test-output/guided-tree_nsamples_%d_nrepetitions_%d_#%d.gv' % (nsamples, nrepetitions, index + 1), view=False)
        dot.format='PNG'
        dot.render('test-output/guided-tree_nsamples_%d_nrepetitions_%d_#%d.gv' % (nsamples, nrepetitions, index + 1), view=False)

# ## True confidence intervals
#
# In the previous sections, we have evaluated the confidence intervals which can be computed on a single sample
# (consisting of multiple runs of applying a guiding strategy). In this section, we compute confidence intervals
# based on multiple repetitions of running a guiding strategy and then comparing the precision of the obtained
# estimate using a fixed amount of simulation runs.

# In[19]:

def compare_strategies_multiple_runs(strategies=None, nrepetitions=1000, nruns=3000, ax=None):
    # number of repetitions to compare guiding strategies against each other
    #nrepetitions = 1000

    # number of simulation runs for each repetition
    #nruns = [10, 100, 1000, 3000]

    #from parallel import mapp
    from shmarray import zeros

    saveStates = {}
    debugStates = []
    saveEtas = []

    #for nr in nruns:
    data = []
    data2 = []

    for j, strat in enumerate(strategies):
        phats = zeros(nrepetitions)
        eta_histogram = zeros((len(strategies), nrepetitions, len(eta)))
        dists_to_hole = zeros((len(strategies), nrepetitions, nruns))
        critn = critais = critz = 0
        loppn = logpz = logpais =0
        logqn = logqz =logqais =0
        #def f(i,saveStates):
        def f(i):
            fvalues, lp, lq, state, leta, dhbox = guided_simulation(nsamples=nruns, strategy=strat)
            phat = estimate_rare_event_prop(fvalues, lp, lq, threshold=1e-09) + 1e-30
            phats[i] = phat
            eta_histogram[j, i,: ] = leta
            dists_to_hole[j, i,: ] = dhbox
           
        mapp(f, range(nrepetitions))
        phats = np.array(phats)



        for i, p in enumerate(phats):
            data.append({'Strategy': strat['name'],
                         'Estimates': cp(p),
                         'choice histogram': np.sum(np.array(eta_histogram)[j,:,:], axis=0)/float(np.sum((np.array(eta_histogram)[j,:,:]).flatten()))
                         })
            data2.append({'Strategy': strat['name'],
                          'dists_to_hole': np.array2string(np.array(dists_to_hole)[j, i, :], separator=",")
                          })




        print("=="*40)
        print("Strategy: "+strat['name'])
        print("Mean estimate:" + str(np.mean(np.array(phats))))
        print("Min-max: " + str(np.min(np.array(phats)))+  ' '+ str(np.max(np.array(phats))))
        print("stdeviation: " + str(np.sqrt(np.var(np.array(phats)))))
        print("Histogram: ", data[-1]['choice histogram'])
    data = pd.DataFrame(data)

    data2 = pd.DataFrame(data2)
    data2.to_csv('distances_2hole_nsamples_%d_nrepetitions_%d.csv' % (nsamples, nreps))

    sns.set_context('talk')
    sns.set_style('ticks')
    fig, ax = pl.subplots()
    ax.set_yscale('log')
    sns.boxplot(x='Strategy', y='Estimates', data=data, ax=ax)
    ax.set_title('Number of simulation runs: %d' % (nruns))
    ax.hlines(y=np.exp(rare_event_log_prob), xmin=ax.get_xlim()[0], xmax=ax.get_xlim()[1],
              label='True rare-event probability', lw=3)
    ax.set_ylim(1e-08, 0.02)
    pl.legend()
    ax.set_yscale('log')
    sns.despine(offset=10)
    fig.savefig('./empirical_confidence_nsamples_%d_nrepetitions_%d_boxplot.pdf' % (nruns,nrepetitions), bbox_inches='tight')

    print('Summary statistics:\n' + str(data.groupby('Strategy').describe()))

    return data


if __name__ == "__main__":
    nsamples = 1000
    nreps = 10000

    # read parameters from commandline
    nsamples = int(sys.argv[1])
    nreps = int(sys.argv[2])
    print("nsamples: " + str(nsamples))
    print("nreps: " + str(nreps))

    strategies = [naive_strategy, init_z_strategy()]
    for b in [10]:
       for q in [0.1]:
           strategies.append(init_cross_entropy_independent_strategy(batchsize=b, quantile=q))

    data=compare_strategies_multiple_runs(strategies, nrepetitions=nreps, nruns=nsamples)
    data.to_csv('evaluation_results_nsamples_%d_nrepetitions_%d.csv'%(nsamples,nreps))
    data.to_pickle('evaluation_results_nsamples_%d_nrepetitions_%d.pckl' % (nsamples, nreps))
